<?php

function site_update_dump($path, $info = array()) {
  $out = fopen(DRUPAL_ROOT . '/' . $path, 'w');

  // Output the PHP header.
  $output = '';
  $output .= "<?php\n";

  $output .= "/**\n";
  $output .= " * @file\n";
  $output .= " * This file was generated by the site_update dump tool.\n";
  $output .= " * Allow the site_update.module to read and write this file.\n";
  $output .= " * Do not edit manually.\n";
  $output .= " *\n";
  $output .= " */\n\n";

  $info += array(
    'version' => SITE_UPDATE_FORMAT_VERSION,
    'time' => REQUEST_TIME,
    'date' => format_date(REQUEST_TIME),
    'base' => url('', array('absolute' => TRUE)),
    'base_id_bump' => SITE_UPDATE_ID_BUMP,
    'base_id_offset' => SITE_UPDATE_ID_OFFSET,
    'base_id_increment' => SITE_UPDATE_ID_INCREMENT,
  );
  $output .= "/** About this database dump. **/\n";
  $output .= "function site_update_dump_info() {\n";
  $output .= "  return " . site_update_var_export($info, '  ') . ";\n";
  $output .= "}\n\n";


  // Enable modules and themes.
  $result = db_query("SELECT * FROM {system} WHERE status = 1 AND type = 'module' ORDER BY weight ASC, name ASC");
  $output .= "/** Enabled modules and their weights. **/\n";
  $output .= "function site_update_dump_modules() {\n";
  $output .= "  return array(\n";
  foreach ($result as $record) {
    $output .= "    '$record->name' => $record->weight,\n";
  }
  $output .= "  );\n";
  $output .= "}\n\n";

  $result = db_query("SELECT * FROM {system} WHERE status = 1 AND type = 'theme' ORDER BY weight ASC, name ASC");
  $output .= "function site_update_dump_themes() {\n";
  $output .= "  return array(\n";
  foreach ($result as $record) {
    $output .= "    '$record->name' => $record->weight,\n";
  }
  $output .= "  );\n";
  $output .= "}\n";

  fwrite($out, $output);
  fflush($out);
  $output = '';

  // tables whose data will not be saved
  $ignore_tables = array(
    'accesslog',
    'advanced_help_index',
    'captcha_sessions',
    'flood',
    'history',
    'menu_router', // rebuilt with menu_rebuild().

    // Registry should be built automatically.
    'registry',
    'registry_file',

    'search_dataset',
    'search_index',
    'search_total',
    'sessions',
    'sequences',
    'site_update_check',
    'system',
    'views_object_cache',
    'watchdog',
  );

  // Allow for site-specific tables to be ignored.
  if (!empty($conf) && !empty($conf['site_update_ignore_tables'])) {
    $ignore_tables = array_merge($ignore_tables, $conf['site_update_ignore_tables']);
  }



  // Get the current schema.
  $schema = drupal_get_schema();
  // Alphabetical order.
  ksort($schema);

  $dumped_tables = array();

  // Export all the tables in the schema.
  foreach ($schema as $table => $data) {
    $data += array(
      // Avoid php notices.
      'description' => '',
      'primary key' => NULL,
    );

    if (empty($data['primary key']) && empty($data['unique keys'])) {
      $msg = t("Site_update dump omitting table !table because it has no unique key.", array('!table' => $table));
      // @todo display msg only if data in table.
      drupal_set_message($msg, 'warning');
      $output .= "// $msg\n\n";
      // $output .= "/* \n" . print_r($data, 1);    $output .= "*/\n\n"; // debug
      continue;
    }

    // Don't output values for those tables.
    if (substr($table, 0, 5) == 'cache' || in_array($table, $ignore_tables)) {
      $output .= "// Omitting table: $table\n\n";
      continue;
    }

    $dumped_tables[] = $table;
    $output .= "/**\n";
    $output .= " * Table {{$table}} provided by {$data['module']} module.\n";
    $output .= " * {$data['description']}\n";
    $output .= " **/\n";

    /* Not needed because fields are same as data keys.
       $output .= "function site_update_dump_fields_$table() {\n";
       $output .= "  return " . site_update_var_export(array_keys($data['fields']), '    ') . ";\n";

       $output .= "} // End table fields: $table\n\n";
    */


    $output .= "function site_update_dump_data_$table() {\n";
    $output .= "  \${$table} = array();\n\n";
    // Prepare the export of values.
    $result = db_query('SELECT * FROM {'. $table .'}', array(), array('fetch' => PDO::FETCH_ASSOC));
    foreach ($result as $record) {
      if (!empty($data['primary key'])) {
        $val = array();
        foreach ($data['primary key'] as $key) {
          $val[$key] = $record[$key];
        }
        // @todo this logic may be unsafe if key values contain '-' or "'" characters!
        $i = implode('-', $val);
      }
      else {
        $i = '';
      }

      // Some rows have table-specific export.
      $function = 'site_update_row_export_' . $table;
      if (!function_exists($function)) {
        $function = 'site_update_var_export';
      }
      $comment = FALSE;
      $export = $function($record, '  ', $comment);
      if ($comment) {
        $output .= "\n  // $comment\n";
      }
      if ($export) {
        $output .= "\n  \${$table}['$i'] = " . $export . ";\n";
      }
      elseif ($i && !$comment) {
        $output .= "\n  // \${$table}['$i'] ommitted from dump.\n";
      }
    }

    $output .= "\n  return \${$table};\n";
    $output .= "} // End table data: $table\n";

    $output .= "\n\n";

    fwrite($out, $output);
    fflush($out);
    $output = '';

  }

  // The list of all tables dumped.
  $output .= "function site_update_dump_tables() {\n";
  $output .= "  return " . site_update_export_simplify(site_update_var_export($dumped_tables, '  ')) . ";\n";
  $output .= "}\n";


  fwrite($out, $output);
  fflush($out);
  $output = '';

  fclose($out);
}

/**
 * Helper function to output insert commands.
 * Most records are inserted as-is.  Some are ommitted from dump.
 */
function site_update_dump_insert_record($table, $record) { // deprecated
  if ($table == 'variable') {
    $matches = preg_filter(site_update_dump_omit_var_regexp(), "$0", $record['name']);
    if (!empty($matches)) {
      return '  // Variable ' . $record['name'] . " not included.\n";
    }
    else {
      return "  site_update_insert('$table', " . site_update_var_export($record, '  ') . ");\n";
    }
  }
  elseif ($table == 'menu_links') {
    //dpm("XXX", __FUNCTION__); // I think this code deprecated!
    // I think safe to omit menu items that have not been customized.
    if (($record['module'] == 'system' || $record['module'] == 'admin_menu') && !$record['customized']) {
      return t('  // Menu link !link_path provided by !module module not customized.', array(
                 '!link_path' => $record['link_path'],
                 '!module' => $record['module'],
               )) . "\n";
    }
    else {
      return "  site_update_insert_menu_link('$table', " . site_update_var_export($record, '  ') . ");\n";
    }
  }
  else {
    return "  site_update_insert('$table', " . site_update_var_export($record, '  ') . ");\n";
  }
}

/**
 * Drupal variables that are never copied from base to working copy.
 */
function site_update_dump_omit_var_regexp() {
  $omit_vars = array(
    '/advanced_help_last_cron/',
    '/bakery_is_master/', // Because the base may be configured as master to simplify login.
    '/cron_key/',
    '/cron_last/',
    '/css_js_query_string/',
    '/ctools_last_cron/',
    '/drupal_private_key/',
    '/drupal_http_request_fails/',
    '/file_private_path/',
    '/file_public_path/',
    '/file_temporary_path/',
    '/file_directory_path/',
    '/install_profile/',
    '/install_task/',
    '/install_time/',
    '/javascript_parsed/',
    '/menu_mask/',
    '/node_cron_last/',
    '/site_name/',
    '/site_mail/',
    '/site_update_hash/',
    '/site_update_is_base/',
  );
  // Allow third party changes.
  drupal_alter('site_update_omit_var_regexp', $omit_vars);
  return $omit_vars;
}

/**
 * Drupal provides drupal_var_export which is supposed to pretty-print
 * code that adheres to Drupal coding standards.  But, it badly
 * mis-handles strings containing single quotes, so it is unusable for our
 * purposes.  This function does not (yet) make the code pretty, but it makes it
 * valid.
 */
function site_update_var_export($var, $prefix = '') {
  // Note this attempt to str_replace the prefix breaks serialized values!
  //return str_replace("\n", "\n$prefix", var_export($var, TRUE));
  return var_export($var, TRUE);
}

/**
 * Omits the numerical indexes of an array.  Desireable to reduce size of diffs and make changes more readable.
 * However, only safe when the var exported is a very simple array structure!  Use caution when calling.
 */
function site_update_export_simplify($text) {
  $text = preg_replace("/\d+ => /", '', $text);
  return $text;
}

function site_update_row_export_variable($record, $prefix = '') {
  $matches = preg_filter(site_update_dump_omit_var_regexp(), "$0", $record['name']);
  if (!empty($matches)) {
    return FALSE;
  }
  else {
    return site_update_var_export($record, $prefix);
  }
}

function site_update_row_export_menu_links($record, $prefix, &$comment) {
  static $customized = array();
  if ($record['customized']) {
    // Customized menu items always exported.
    $customized[$record['mlid']] = TRUE;
  }
  elseif ($record['plid'] && !empty($customized[$record['plid']])) {
    // Descendents of customized items also exported.
    $customized[$record['mlid']] = TRUE;
  }

  if (!empty($customized[$record['mlid']])) {
    return site_update_var_export($record, $prefix);
  }
  else {
    $comment = t('Menu link \'!path\' in !menu menu provided by !module.', array(
                   '!path' => $record['link_path'],
                   '!module' => $record['module'],
                   '!menu' => $record['menu_name'],
                 ));
    return FALSE;
  }
}

function site_update_row_export_block($record, $prefix, &$comment) {
  if (!$record['status'] && !$record['custom']) {
    $comment = strtr('Block module-delta disabled in theme.', $record);
    return FALSE;
  }
  else {
    return site_update_var_export($record, $prefix);
  }
}